#include <Wire.h>
#include <math.h>

//Adresses
#define ACCEL_Y_H       0x3B   // R  
#define ACCEL_Y_L       0x3C   // R  
#define ACCEL_X_H       0x3D   // R  
#define ACCEL_X_L       0x3E   // R  
#define ACCEL_Z_H       0x3F   // R  
#define ACCEL_Z_L       0x40   // R  
#define TEMP_OUT_H      0x41   // R  
#define TEMP_OUT_L      0x42   // R  
#define GYRO_Y_H        0x43   // R  
#define GYRO_Y_L        0x44   // R  
#define GYRO_X_H        0x45   // R  
#define GYRO_X_L        0x46   // R  
#define GYRO_Z_H        0x47   // R  
#define GYRO_Z_L        0x48   // R 
int gyro = 0x68;
int gyro_pwr_m_1 = 0x6B;
int who_am_i = 0x75;

//Pins
int startLED = 6;
int PWMA = 11;
int AIN2 = 13;
int AIN1 = 12;
int STBY = 10;
int BIN1 = 8;
int BIN2 = 7;
int PWMB = 9;

//Global Variables

int accel_x_reading = 0;
int accel_y_reading = 0;
int accel_z_reading = 0;
int gyro_x_reading = 0;
int gyro_y_reading = 0;
int gyro_z_reading = 0;
int accel_pitch = 0;
int accel_vector = 0;
int temp1 = 0;
float temp2 = 0;
int power = 0;
float sinus = 0;

float gyro_pitch = 0;
unsigned long timer = 0;

float pitch = 0;
float corr_pitch = 0;

//Serial revieve varialbles
char inputArray[20];
String inputString = ""; 
boolean stringComplete = false;


//PID Variables 
int pid_dt = 5; //Delay for PID in milliseconds
unsigned long last_PID = 0;
double Pterm = 0.0;
double Pgain = 1.9;
double Iterm = 0.0;
double Istate = 0.0;
double Igain = 0.0;
double Dterm = 0.0;
double Dgain = 0.0;
double last_angle = 0.0;




void setup() {
  
  Serial.begin(9600); 
  inputString.reserve(200);
  Wire.begin();
  
  //Blinking LED to indicate start
  for (int i = 0; i < 3; i++){
    digitalWrite(startLED, HIGH);
    delay(150);
    digitalWrite(startLED, LOW);
    delay(150);
  }

  //Writing a zero to the sleep mode register
  Wire.beginTransmission(gyro);
  Wire.write(gyro_pwr_m_1); //Write adress of the register
  Wire.write(0); //Write data and size of data in bytes
  Wire.endTransmission(true); // send and release the bus
  initializeGyro();
  
  timer = millis();
  
  delay(400);
  digitalWrite(STBY,HIGH);
  
  /*Spin wheels to indicate start
  analogWrite(PWMA,125);
  digitalWrite(AIN1,HIGH);
  digitalWrite(AIN2,LOW);
  analogWrite(PWMB,125);
  digitalWrite(BIN1,HIGH);
  digitalWrite(BIN2,LOW);
    
  delay(200);
    
  digitalWrite(AIN1,LOW);
  digitalWrite(AIN2,HIGH);
  digitalWrite(BIN1,LOW);
  digitalWrite(BIN2,HIGH);
  delay(200);
  */
   
  
}

void loop() {
  
  //delay(3);
  
  accel_pitch = getAccelPitch();
  gyro_pitch = getGyroPitch();
  pitch = (0.95)*gyro_pitch+(0.05)*accel_pitch; //Complimentary filter
  corr_pitch = pitch - 5; //Correction for gyro position on the robot
  
  sendData();
  
  power = pid();
  if (power > 0) {
       digitalWrite(AIN1,HIGH);
       digitalWrite(AIN2,LOW);
       digitalWrite(BIN1,HIGH);
       digitalWrite(BIN2,LOW);
   } 
   else {
       digitalWrite(AIN1,LOW);
       digitalWrite(AIN2,HIGH);
       digitalWrite(BIN1,LOW);
       digitalWrite(BIN2,HIGH);
   }
   
   analogWrite(PWMA,abs(power)); 
   analogWrite(PWMB,abs(power));
   
   
  //Reading temperature
  //temp1 = read_16_bit_register(TEMP_OUT_H);
  //temp2 = (temp1/340.0) + 36.53;

  
}

//SUPPORT FUNCTIONS****************************************************************************************

int read_16_bit_register (int reg) {
  
  Wire.beginTransmission(gyro);
  Wire.write(reg); //Write adress of the register
  Wire.endTransmission(false); // Send, but don't release the connection
  Wire.requestFrom(gyro, 2,true);   // request 1 byte and end the connection (true 
  
  while(Wire.available())    // slave may send less than requested
  { 
    int data1 = Wire.read();    // receive a byte as int
    int data2 = Wire.read();
    return (data2 | data1 << 8);
  }  
  
}

int check_sleep_mode (void) {
  Wire.beginTransmission(gyro);
  Wire.write(gyro_pwr_m_1); //Write adress of the register
  Wire.endTransmission(false); // Send, but don't release the connection
  Wire.requestFrom(gyro, 1,true);   // request 1 byte and end the connection (true 
  int sleep = 0;
  while(Wire.available())    // slave may send less than requested
  { 
    sleep = Wire.read();    // receive a byte as int
    Serial.print(("  Sleep = "));
    Serial.println(sleep);   
  }
  
  if(sleep == 0){
    return (0);
  } 
  else{
    return (1);
  }

} 

void sendData (void) {
    Serial.print('A');
    Serial.println(accel_pitch);
    Serial.print('B');
    Serial.println(Pgain*1000);
    Serial.print('C');
    Serial.println(power);
    Serial.print('D');
    Serial.println(Igain*1000);
    Serial.print('E');
    Serial.println(corr_pitch);
    Serial.print('F');
    Serial.println(Dgain*1000);
    Serial.print('G');
    Serial.println(accel_vector); 
}  

float getAccelPitch (void) {
  
  float pitch = 0;
  accel_x_reading = read_16_bit_register(ACCEL_X_H);
  accel_y_reading = -read_16_bit_register(ACCEL_Y_H);
  accel_z_reading = read_16_bit_register(ACCEL_Z_H);
  accel_vector = sqrt(pow(accel_x_reading,2) + pow(accel_y_reading,2) + pow(accel_z_reading,2));
  if(accel_y_reading > 5000){
    pitch = 57.2958*acos((float)accel_z_reading/accel_vector) - 90;
  }
  else if(accel_y_reading < 5000  && accel_y_reading > -5000) {
      if(accel_z_reading < 0 ){ 
          pitch = 57.2958*acos((float)accel_y_reading/accel_vector);  
      } 
      else { 
      pitch = (-1)*57.2958*acos((float)accel_y_reading/accel_vector);  
      }
  }
  else{
      if (accel_z_reading > 0){
      pitch = -57.2958*acos((float)accel_z_reading/accel_vector) -90.0;
      }
      else{
      pitch = -57.2958*acos((float)accel_z_reading/accel_vector) +270.0;
      }
  }
  return(pitch);
}


float getGyroPitch (){
  float angle = pitch;
  gyro_x_reading = read_16_bit_register(GYRO_X_H);
  gyro_y_reading = read_16_bit_register(GYRO_Y_H);
  gyro_z_reading = read_16_bit_register(GYRO_Z_H);
  if (gyro_x_reading < -1500 || gyro_x_reading > 1500){
        angle += ((double)gyro_x_reading*(micros()-timer))/131000000.0;
  }
  if (angle > 130 && accel_pitch < 0){
      angle = angle - 360;
  }
  else if (angle < -130 && accel_pitch > 0){
      angle = angle + 360;
  }
  timer = micros();
  
  return(angle); 
  
}

void initializeGyro (void) {
  while(shaking()){}
  gyro_pitch = getAccelPitch();
}


int shaking(void){
  getAccelPitch();
  if (accel_vector > 19600 || accel_vector < 13100){
      return(1);
  }
  else{
      return(0);
  }
}

//********************Setup PID loop to determine torque************************* 
int pid() { 
  if((millis()-last_PID) >= pid_dt){      // sample every dt ms -> 1000/dt hz.
    last_PID = millis();
    Pterm = corr_pitch * Pgain;
    Istate = (Istate + corr_pitch);
    Iterm = Istate * Igain;
    Dterm = (corr_pitch - last_angle) * Dgain;
    int vTorque = Pterm - Dterm + Iterm;
    last_angle = corr_pitch;
    vTorque = constrain(vTorque,-255,255);
    return(vTorque);
  }
}
//*************End Setup PID Loop************************************************


//Serial event to recieve data over serial
void serialEvent() {
  Serial.println("Serial event!!");
  while (Serial.available()) {
    char inChar = (char)Serial.read(); 
    if (inChar == '$') {
      stringComplete = true;
    }else{ 
      inputString += inChar;
    }
   }
   
   if(stringComplete){
      
      String numberString = inputString.substring(2,7);
      
      numberString.toCharArray(inputArray,20);
      long inputToLong = atol(inputArray);
      if(inputString.charAt(0) == 'P' && inputString.charAt(1) == 'V'){
        Pgain = inputToLong/1000.0;
      }  
      else if(inputString.charAt(0) == 'I' && inputString.charAt(1) == 'V'){
        Igain = inputToLong/1000.0; 
      } 
      else if(inputString.charAt(0) == 'D' && inputString.charAt(1) == 'V'){
        Dgain = inputToLong/1000.0; 
      }
      inputString = "";
      stringComplete = false;
    }
}
